from abc import ABCMeta, abstractmethod


class BinaryBase(metaclass=ABCMeta):
    def __init__(self, name, offsets=None):
        """
        Constructor.

        :param name: the binary's name
        :type name: str
        :param offsets: initial offsets that are stored into the binary
        :type offsets:  dict or None
        """
        self._name = name
        if offsets:
            self._offsets = offsets.copy()
        else:
            self._offsets = dict()

    def get_name(self):
        """
        Returns the name.
        """
        return self._name

    @abstractmethod
    def get_address(self, name, offset=0):
        """
        Resolves the address of the symbol with the given name and adds the
        given offset.

        :param name: the symbol's name
        :type name: str
        :param offset: an offset that is added to the returned address
        :type offset: int
        """
        pass

    def __setitem__(self, name, offset):
        """
        Sets the offset of the symbol with the given name.

        :param name: the symbol's name
        :type name: str
        :param offset: the offset of the symbol within the binary
        :type offset: int
        """
        self.add_offset(name, offset)

    def __getitem__(self, name):
        """
        Resolves the address of the symbol with the given name.

        :param name: the symbol's name
        :type name: str
        :raises KeyError:
        """
        address = self._get_address(name, 0)
        if address is None:
            raise KeyError("'{}' cannot be looked up in {}"
                           .format(name, self._name))
        return address

    def add_offset(self, name, offset):
        """
        Adds the offset of the symbol with the given name.

        :param name: the symbol's name
        :type name: str
        :param offset: the offset of the symbol within the binary
        :type offset: int
        """
        if name in self._offsets:
            raise ValueError("Duplicated offset '{}'".format(name))
        self._offsets[name] = offset



from .util import encode_as_bytes, pack_value


class RopChain:
    def __init__(self, pwny):
        """Constructor.

        :param pwny.Pwny pwny: the parent Pwny that has created the chain
        """
        self._pwny = pwny
        self._chain = list()

    def get_pwny(self):
        """
        Returns the parent pwny.
        """
        return self._pwny

    def append(self, obj):
        """Appends a generic object to the rop chain and tries to convert is to
        byte string.

        Long.

        :param object obj: the generic object that is to be added
        """
        self._chain.append(obj)

    def pad(self, length, scale=1, value=b"_"):
        """
        Pads the chain with length multiplied by scale times the given value.
        """
        if isinstance(value, int):
            value = pack_value(value, 1)
        self._chain.append(length * scale * value)

    def byte(self, value):
        """
        Adds a byte to the chain.
        """
        self.value(value, size=1)

    def newline(self):
        """
        Adds a newline to the chain.
        """
        self.byte(0xa)

    def word(self, value, little_endian=None):
        """
        Adds a word (2 bytes) to the chain.
        """
        self.value(value, size=2, little_endian=little_endian)

    def dword(self, value, little_endian=None):
        """
        Adds a double word (4 bytes) to the chain.
        """
        self.value(value, size=4, little_endian=little_endian)

    def qword(self, value, little_endian=None):
        """
        Adds a quad word (8 bytes) to the chain.
        """
        self.value(value, size=8, little_endian=little_endian)

    def value(self, value, size=None, little_endian=None):
        """
        Adds a value to the chain. The value is packed into the given size
        bytes with the given endianess. If none is specified the word size and
        endianess of the parent Pwny is used.
        """
        self._chain.append(self._pack_value(value, size, little_endian))

    def addr(self, name, bin_name=None, size=None, little_endian=None):
        self._chain.append(lambda: self._pack_addr(name, bin_name, size,
                                                   little_endian))

    def _pack_addr(self, name, bin_name, size, little_endian):
        address = self._pwny._binary_store.get_address(name, bin_name)
        return self._pack_value(address, size, little_endian)

    def _pack_value(self, value, size, little_endian):
        size = size or self._pwny._word_size
        little_endian = little_endian or self._pwny._little_endian
        return pack_value(value, size, little_endian)

    def serialize(self):
        """
        Serializes the chain into a byte string.
        """
        serialized = b""
        for part in self._chain:
            if callable(part):
                part = part()
            serialized += encode_as_bytes(part)
        return serialized

    def send(self):
        """
        Sends the serialized chain to the remote host using the parent pwny.
        """
        self._pwny.send(self.serialize())



from .binarybase import BinaryBase


class StaticBinary(BinaryBase):
    def __init__(self, name, offsets=None):
        super().__init__(name, offsets)

    def get_address(self, name, offset=0):
        address_offset = self._offsets.get(name)
        if address_offset is None:
            return None
        return address_offset + offset



import itertools
import struct
from functools import wraps


WORD_SIZE_FORMAT_LOOKUP = {
    1: "B",
    2: "H",
    4: "I",
    8: "Q"
}


def pack_value(value, size=8, little_endian=True):
    pack_format = "<"
    if not little_endian:
        pack_format = ">"
    pack_format += WORD_SIZE_FORMAT_LOOKUP[size]
    return struct.pack(pack_format, value)


def encode_as_bytes(data):
    if isinstance(data, bytes):
        return data
    if not isinstance(data, str):
        data = str(data)
    return data.encode()


def encode_arg(index):
    def _encode_arg(f):
        def _inner(self, *args, **kwargs):
            before = args[0:index]
            arg = args[index]
            after = args[index+1:]
            if not isinstance(arg, bytes):
                arg = encode_as_bytes(arg)
            args = list(itertools.chain(before, [arg], after))
            return f(self, *args, **kwargs)
        return _inner
    return _encode_arg


def catch_all(f):
    return catch_all_with_error()(f)


def catch_all_with_error(error="An error occurred: ", print_exception=True):
    def _catch_all_with_error(f):
        @wraps(f)
        def _inner(*args, **kwargs):
            try:
                f(*args, **kwargs)
            except Exception as e:
                if print_exception:
                    print("{}: {}".format(error, e))
                else:
                    print(error)
        return _inner
    return _catch_all_with_error



from struct import pack, unpack


def _get_format(form, little_endian):
    if little_endian:
        return "<" + form
    return ">" + form


def p8(value):
    return pack("B", value)


def p16(value, little_endian=True):
    return pack(_get_format("H", little_endian), value)


def p32(value, little_endian=True):
    return pack(_get_format("I", little_endian), value)


def p64(value, little_endian=True):
    return pack(_get_format("Q", little_endian), value)


def u8(value):
    return unpack("B", value)


def u16(value, little_endian=True):
    return unpack(_get_format("H", little_endian), value)


def u32(value, little_endian=True):
    return unpack(_get_format("I", little_endian), value)


def u64(value, little_endian=True):
    return unpack(_get_format("Q", little_endian), value)



def parse_pointer_leak(leaked_data):
    if isinstance(leaked_data, bytes):
        leaked_data = leaked_data.decode()

    leaked_data = leaked_data.replace("(nil)", "0x0")
    return list(map(lambda addr: int(addr, 16), parts))



import itertools
import socket
import telnetlib
import time
from functools import wraps
from .binarystore import BinaryStore
from .ropchain import RopChain
from .util import encode_arg


def delayed(f):
    """
    Decorates an instance method of a class and delays its execution by the
    value of the instances internal _delay field.
    """
    @wraps(f)
    def _inner(self, *args, **kwargs):
        time.sleep(self._delay)
        return f(self, *args, **kwargs)
    return _inner


class Pwny:
    def __init__(self, sock=None, dump_received=True, try_decode_received=True,
                 word_size=8, little_endian=True, delay=0.1,
                 confirm_send=False, binaries=None):
        """
        Constructor.

        :param socket.socket sock: a initialized non-connected socket or None if a socket
            shall be created
        """
        self._sock = sock or socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._dump_received = dump_received
        self._try_decode_received = try_decode_received
        self._word_size = word_size
        self._little_endian = little_endian
        self._delay = delay
        self._confirm_send = confirm_send
        self._binary_store = BinaryStore()
        binaries = binaries or list()
        for binary in binaries:
            self._binary_store.add_binary(binary)

    def connect(self, host, port):
        """
        Connects to the given host and port.
        """
        self._sock.connect((host, port))

    def half_close(self):
        """
        Half closes the internal socket so that a remote read will fail but
        remote messages can still be received.
        """
        self._sock.shutdown(socket.SHUT_WR)

    def close(self):
        """
        Closes the internal socket.
        """
        self._sock.close()

    def get_socket(self):
        """
        Returns the internal socket.
        """
        return self._sock

    def set_little_endian(self, little_endian):
        """
        Sets if the exploit assumes a little or big endian system.
        """
        self._little_endian = little_endian

    def set_word_size(self, word_size):
        """Sets the word size of the exploited binary.

        :param int word_wize: must be 1, 2, 4 or 8
        """
        if word_size not in [1,2,4,8]:
            raise Exception("Illegal word size: %d".format(word_size))
        self._word_size = word_size

    def set_dump_received(self, dump_received):
        """
        Sets whether received data should be dumped to stdout.
        """
        self._dump_received = dump_received

    def set_try_decode_received(self, try_decode_received):
        """
        Sets whether received binary data should be tried to be decoded.
        """
        self._try_decode_received = try_decode_received

    def set_confirm_send(self, confirm_send):
        """
        Sets whether the user must confirm sending data by pressing enter.
        """
        self._confirm_send = confirm_send

    def set_delay(self, delay):
        """
        Sets the delay between single send and/or received operations.

        :param delay: the delay in seconds
        :type delay: float
        """
        self._delay = delay

    def add_binary(self, binary, bin_name=None):
        """
        Adds a binary to the internal binary store, that can be used to resolve
        symbols to addresses.

        :param binary: the binary that is to be added
        :type binary: BinaryBase and its subclasses
        :param bin_name: the name of the binary, if None the binaries internal
            name is used
        :type bin_name: str or None
        """
        self._binary_store.add_binary(binary, bin_name)

    @delayed
    @encode_arg(0)
    def send(self, data):
        """
        Tries to encode the given data as bytes and send it to the remote.

        :param data: the data that is send to the remote
        :type data: bytes or str
        """
        if self._confirm_send:
            self.prompt("Press enter to send:\n{}".format(data))
        self._sock.sendall(data)

    @delayed
    def recv(self, buffer_size=4096):
        """
        Receives data from the remote.

        :param buffer_size: the size of the receive buffer and thus the maximum
            amount of data that is received at once
        :type buffer_size: int
        """
        received = self._sock.recv(buffer_size)
        self._try_decode_and_print(received)
        return received

    @delayed
    def recv_length(self, length):
        """
        Receives the given amount of bytes from the remote.

        :param length: the length that should be received in bytes
        :type length: int
        """
        received = b""
        while len(received) < length:
            received += self._sock.recv(length - len(received))
        self._try_decode_and_print(received)
        return received

    @delayed
    @encode_arg(0)
    def recv_expect(self, expected):
        """
        Receives the expected string from the remote.

        :param expected: the string that is expected to be received from the
            remote
        :type expected: bytes or str
        """
        received = b""
        while received != expected:
            received += self._sock.recv(1)
        self._try_decode_and_print(received)
        return received

    @delayed
    @encode_arg(0)
    def recv_until(self, until):
        """
        Receives until the received data contains the given string.

        :param until: the string that is expected to be received eventually
            from the remote
        :type until: bytes or str
        """
        received = b""
        while until not in received:
            received += self._sock.recv(1)
        self._try_decode_and_print(received)
        return received

    def recv_line(self):
        """
        Receives from the remote until a newline is received.
        """
        return self.recv_until(b"\n")

    def recv_lines(self, count):
        for _ in range(count):
            self.recv_line(self)

    def prompt(self, prompt=None):
        """
        Prompts the user for some input.

        :param prompt: the prompt that is displayed to the user
        :type prompt: str
        """
        prompt = prompt or ""
        return input(prompt)

    def interact(self):
        """
        Lets the user interactively communicate with the remote.
        """
        t = telnetlib.Telnet()
        t.sock = self._sock
        t.interact()

    def _try_decode_and_print(self, data):
        """
        Prints the given data to stdout if self._dump_received is True. If
        self._try_decode_received is True, the given data will be tried to be
        decoded. If this does not work the raw data will be printed.
        """
        if not self._dump_received:
            return
        if self._try_decode_received:
            try:
                print(data.decode(), end="")
            except:
                print(data)
        else:
            print(data)

    def create_chain(self):
        """
        Creates a new rop chain.
        """
        return RopChain(self)

    @staticmethod
    def from_socket(sock):
        """
        Creates a new Pwny that wraps the given socket.

        :param sock: that socket that is to be wrapped
        :type sock: socket.socket
        """
        return Pwny(sock)



from .binarystore import BinaryStore
from .dynamicbinary import DynamicBinary
from .pwny import Pwny
from .ropchain import RopChain
from .staticbinary import StaticBinary
from .util import catch_all, catch_all_with_error



from distutils.core import setup


setup(
    name="pwny",
    version="0.1.4",
    author="André Kupka",
    author_email="kupka@in.tum.de",
    description="A small binary exploitation framework.",
    license="MIT",
    keywords="binary exploitation pwning pwny",
    url="https://github.com/fr3akout/pwny",
    packages=["pwny", "pwny.formatstring", "pwny.util"],
    long_description="A small binary exploitation framework, that provides "
            "functionality to communicate with remote servers and write rop "
            "chains.",
    classifiers=[]
)



from .binarybase import BinaryBase


class DynamicBinary(BinaryBase):
    def __init__(self, name, offsets=None, base_address=None):
        """
        Constructor.

        :param name: the binary's name
        :type name: str
        :param offsets: initial offsets that are stored into the binary
        :type offsets: dict or None
        :param base_address: the binary's base address in virtual memory
        :type base_address: int or None
        """
        super().__init__(name, offsets)
        self._base_address = base_address

    def set_base_address(self, base_address):
        """
        Sets the base address of the binary. The base address is the start
        address where this binary has been mapped into virtual memory.

        :param base_address: the base address
        :type base_address: int
        """
        self._base_address = base_address

    def get_address(self, name, offset=0):
        """
        TODO inherit doc
        """
        if self._base_address is None:
            raise Exception("no base address has been set")
        address_offset = self._offsets.get(name)
        if address_offset is None:
            return None
        return address_offset + self._base_address + offset



from collections import OrderedDict
from .dynamicbinary import DynamicBinary


class BinaryStore:
    def __init__(self):
        self._binaries = OrderedDict()

    def add_binary(self, binary, name=None):
        if name is None:
            name = binary.get_name()
        if name in self._binaries:
            raise ValueError("Duplicated binary '{}'".format(name))
        self._binaries[name] = binary

    def get_binary(self, bin_name):
        return self._binaries.get(bin_name)

    def set_base_address(self, bin_name, base_address):
        binary = self._binaries[bin_name]
        if not isinstance(binary, DynamicBinary):
            raise ValueError("Cannot set base address for non-dynamic binary.")
        binary.set_base_address(base_address)

    def get_address(self, name, bin_name=None, offset=0):
        if bin_name is None:
            return self._search_all_binaries(name, offset)
        binary = self._binaries.get(bin_name)
        if binary is None:
            raise ValueError("No binary with name '{}' exists".format(bin_name))
        address = binary.get_address(name, offset)
        if address is None:
            raise ValueError("No symbol with name '{}' exists".format(name))
        return address

    def _search_all_binaries(self, name, offset):
        for binary in self._binaries.values():
            address = binary.get_address(name, offset)
            if address is not None:
                return address
        raise ValueError("No symbol with name '{}' exists".format(name))
